from __future__ import annotations

import random
from collections import defaultdict
from itertools import groupby
from pathlib import Path

import midsv
import pysam

from DAJIN2.utils import sam_handler


def realign(sam: list[list[str]], GENOME_COODINATES: dict) -> list[str]:
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@")]
    for s in sam_headers:
        if s[0] != "@SQ":
            continue
        s[1] = f'SN:{GENOME_COODINATES["chrom"]}'
        s[2] = f'LN:{GENOME_COODINATES["chrom_size"]}'
    for s in sam_contents:
        s[2] = GENOME_COODINATES["chrom"]
    if GENOME_COODINATES["strand"] == "-":
        sam_contents = sam_handler.revcomp_sam(sam_contents, GENOME_COODINATES["end"])
    else:
        for s in sam_contents:
            s[3] = str(int(s[3]) + GENOME_COODINATES["start"] - 1)
    return sam_headers + sam_contents


def group_by_name(sam_contents: list[str], clust_sample: list[dict]) -> dict[list]:
    sam_contents.sort()
    clust_sample_qname = sorted(clust_sample, key=lambda x: x["QNAME"])
    clust_sample_qname_set = set()
    for qnames in clust_sample_qname:
        qname = qnames["QNAME"]
        clust_sample_qname_set.add(qname)
    sam_groups = defaultdict(list)
    idx_left = 0
    idx_right = 0
    while idx_left < len(sam_contents) and idx_right < len(clust_sample_qname):
        read_left = sam_contents[idx_left][:-1]
        read_right = clust_sample_qname[idx_right]
        qname_left = read_left[0]
        qname_right = read_right["QNAME"]
        if qname_left not in clust_sample_qname_set:
            idx_left += 1
            continue
        if qname_left == qname_right:
            key = read_right["NAME"]
            sam_groups[key].append(read_left)
            idx_left += 1
        else:
            idx_right += 1
    return sam_groups


###############################################################################
# igvjs
###############################################################################


def subset_qnames(RESULT_SAMPLE, readnum: int = 100) -> dict[set[str]]:
    qnames_by_name = defaultdict(set)
    for name, group in groupby(RESULT_SAMPLE, key=lambda x: x["NAME"]):
        group = list(group)
        qnames = [res["QNAME"] for res in group[:readnum]]
        qnames_by_name[name] = set(qnames)
    return qnames_by_name


def subset_reads(name, sam_content, qnames_by_name):
    qnames = qnames_by_name[name]
    sam_subset = [sam for sam in sam_content if sam[0] in qnames]
    return sam_subset


###############################################################################
# Output
###############################################################################


def write_sam_to_bam(sam: list[list[str]], path_sam: str | Path, path_bam: str | Path, threads: int = 1) -> None:
    formatted_sam = "\n".join("\t".join(s) for s in sam)
    Path(path_sam).write_text(formatted_sam + "\n")
    pysam.sort("-@", f"{threads}", "-o", str(path_bam), str(path_sam))
    pysam.index("-@", f"{threads}", str(path_bam))


def update_sam(sam: list, GENOME_COODINATES: dict = None) -> list:
    sam_update = sam.copy()
    sam_update = sam_handler.remove_overlapped_reads(sam_update)
    sam_update = sam_handler.remove_microhomology(sam_update)
    if GENOME_COODINATES["genome"]:
        sam_update = realign(sam_update, GENOME_COODINATES)
    return sam_update


def output_bam(TEMPDIR, NAME, GENOME_COODINATES, THREADS, RESULT_SAMPLE=None, is_control=False) -> None:
    randomnum = random.randint(100_000, 999_999)
    path_sam_input = Path(TEMPDIR, NAME, "sam", "map-ont_control.sam")
    sam = list(midsv.read_sam(path_sam_input))
    # Update sam
    sam_update = update_sam(sam, GENOME_COODINATES)
    # Output SAM and BAM
    path_sam_output = Path(TEMPDIR, "report", "BAM", f"tmp{randomnum}_{NAME}_control.sam")
    path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}.bam")
    write_sam_to_bam(sam_update, path_sam_output, path_bam_output, THREADS)
    # Prepare SAM headers and contents
    sam_headers = [s for s in sam_update if s[0].startswith("@")]
    sam_contents = [s for s in sam_update if not s[0].startswith("@")]
    if is_control:
        qnames = set(list(set(s[0] for s in sam_contents[:10000]))[:100])
        sam_subset = [s for s in sam_update if s[0] in qnames]
        path_sam_output = Path(TEMPDIR, "report", "BAM", f"tmp{randomnum}_{NAME}_control_cache.sam")
        path_bam_output = Path(TEMPDIR, "cache", ".igvjs", NAME, "control.bam")
        write_sam_to_bam(sam_headers + sam_subset, path_sam_output, path_bam_output, THREADS)
    else:
        sam_groups = group_by_name(sam_contents, RESULT_SAMPLE)
        qnames_by_name = subset_qnames(RESULT_SAMPLE)
        # Output SAM and BAM
        for name, sam_content in sam_groups.items():
            # BAM
            path_sam_output = Path(TEMPDIR, "report", "bam", f"tmp{randomnum}_{name}.sam")
            path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}_{name}.bam")
            write_sam_to_bam(sam_headers + sam_content, path_sam_output, path_bam_output, THREADS)
            # igvjs
            sam_subset = subset_reads(name, sam_content, qnames_by_name)
            path_sam_output = Path(TEMPDIR, "report", "bam", f"tmp{randomnum}_{name}_subset.sam")
            path_bam_output = Path(TEMPDIR, "report", ".igvjs", NAME, f"{name}.bam")
            write_sam_to_bam(sam_headers + sam_subset, path_sam_output, path_bam_output, THREADS)
    # Remove temporary files
    sam_temp = Path(TEMPDIR, "report", "BAM").glob(f"tmp{randomnum}*.sam")
    [s.unlink() for s in sam_temp]



from __future__ import annotations

import midsv
import pickle
from pathlib import Path
from DAJIN2.core.preprocess import mapping


def select_control(path_fasta: list[Path]) -> Path:
    return [f for f in path_fasta if f.stem == "control"][0]


def select_others(path_fasta: list[Path]) -> list[Path]:
    return [f for f in path_fasta if f.stem != "control"]


def has_splice(alignments_midsv: dict[str]) -> bool:
    cssplits = alignments_midsv["CSSPLIT"].split(",")
    i = 0
    while i < len(cssplits):
        if cssplits[i].startswith("+") and (cssplits[i + 1].startswith("-") or cssplits[i + 1].startswith("N")):
            length_ins = cssplits[i].count("+")
            current_op = cssplits[i + 1][0]
            length_del = 0
            while current_op == cssplits[i + 1][0]:
                length_del += 1
                i += 1
            if length_ins == length_del:
                return True
        i += 1
    return False


def get_cssplit(reference: Path, query: Path, preset: str = "splice") -> list[str]:
    alignments = [a.split("\t") for a in mapping.to_sam(reference, query, preset=preset)]
    alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
    if has_splice(alignments_midsv):
        alignments = [a.split("\t") for a in mapping.to_sam(reference, query, preset="map-ont")]
        alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
    return alignments_midsv["CSSPLIT"].split(",")


def calculate_index_mapping(cssplits: list[str]) -> dict[int, int]:
    index_mapping = dict()
    query_index = 0
    range_inversion = set()
    for reference_index, element in enumerate(cssplits):
        # Inversion: If the last character is lower, it means we're in an inverted range
        if reference_index in range_inversion:
            continue
        if element[-1].islower():
            start = reference_index
            while element[-1].islower() and reference_index < len(cssplits):
                end = reference_index
                range_inversion.add(reference_index)
                reference_index += 1
                query_index += 1
                element = cssplits[reference_index]
            query_indexes = {i: j for i, j in zip(range(start, end + 1), range(end, start - 1, -1))}
            index_mapping.update(query_indexes)
        # Deletion or Unknown: Skip incrementing query_index
        elif element.startswith("-") or element == "N":
            continue
        # Substition: Just incrementing query_index
        elif element.startswith("*"):
            query_index += 1
        # Insertion: incrementing query_index by the number of insertions
        elif element.startswith("+"):
            query_index += element.count("+")
            index_mapping[reference_index] = query_index
            query_index += 1
        else:
            index_mapping[reference_index] = query_index
            query_index += 1
    return index_mapping


def get_index_mapping(TEMPDIR: str | Path) -> dict[str, dict[int, int]]:
    """
    Returns:
        dict(set): corresponding index of each allele based on control allele
        key: index of control allele
        value: index of other allele
    """
    path_fasta = [f for f in Path(TEMPDIR, "fasta").iterdir() if f.suffix != ".fai"]
    reference = select_control(path_fasta)
    INDEX_MAPPING = dict()
    for query in select_others(path_fasta):
        cssplits = get_cssplit(reference, query, preset="splice")
        allele = query.stem
        INDEX_MAPPING[allele] = calculate_index_mapping(cssplits)
    return INDEX_MAPPING


def save_index_mapping(TEMPDIR: str | Path) -> None:
    INDEX_MAPPING = get_index_mapping(TEMPDIR)
    with open(Path(TEMPDIR, "mutation_loci", "index_mapping.pickle"), "wb") as f:
        pickle.dump(INDEX_MAPPING, f)



from __future__ import annotations

from pathlib import Path

import midsv


def _replaceNtoD(cssplits_sample, sequence) -> list[list[str]]:
    cssplits_replaced = cssplits_sample.copy()
    for i, cssplits in enumerate(cssplits_sample):
        # extract right/left index of the end of sequential Ns
        left_idx_n = 0
        for cs in cssplits:
            if cs != "N":
                break
            left_idx_n += 1
        right_idx_n = 0
        for cs in cssplits[::-1]:
            if cs != "N":
                break
            right_idx_n += 1
        right_idx_n = len(cssplits) - right_idx_n - 1
        # replace sequential Ns within the sequence
        for j, (cs, seq) in enumerate(zip(cssplits, sequence)):
            if left_idx_n <= j <= right_idx_n and cs == "N":
                cssplits_replaced[i][j] = f"-{seq}"
    return cssplits_replaced


###############################################################################
# main
###############################################################################

def replace_NtoD(midsv_sample_alleles: dict[str, list[dict]], FASTA_ALLELES: dict) -> dict[str, list[dict]]:
    """
    Convert any `N` as deletions other than consecutive `N` from both ends
    """
    midsv_updated = dict()
    for allele, sequence in FASTA_ALLELES.items():
        midsv_sample = midsv_sample_alleles[allele]
        cssplits_sample = [cs["CSSPLIT"].split(",") for cs in midsv_sample]
        cssplits_replaced = _replaceNtoD(cssplits_sample, sequence)
        midsv_cssplits = [",".join(cs) for cs in cssplits_replaced]
        # Save as a json
        for i, cssplits in enumerate(midsv_cssplits):
            midsv_sample[i]["CSSPLIT"] = cssplits
        midsv_updated[allele] = midsv_sample
    return midsv_updated



from __future__ import annotations

import midsv

from pathlib import Path
from typing import Generator
from itertools import chain, groupby

from DAJIN2.utils import sam_handler
from DAJIN2.utils import cssplits_handler


def _has_inversion_in_splice(CIGAR: str) -> bool:
    previous_insertion = False
    for cigar in sam_handler.split_cigar(CIGAR):
        if cigar.endswith("I"):
            previous_insertion = True
            continue
        if previous_insertion and cigar.endswith("N"):
            return True
        else:
            previous_insertion = False
    return False


def extract_qname_of_map_ont(sam_ont: Generator[list[str]], sam_splice: Generator[list[str]]) -> set[str]:
    """Extract qname of reads from `map-ont` when:
    - no inversion signal in `splice` alignment (insertion + deletion)
    - single read
    - long alignment length
    """
    alignments_splice = {s[0]: s for s in sam_splice if not s[0].startswith("@")}
    alignments_ont = sorted([s for s in sam_ont if not s[0].startswith("@")], key=lambda x: x[0])
    qname_of_map_ont = set()
    for qname_ont, group in groupby(alignments_ont, key=lambda x: x[0]):
        group = list(group)

        if qname_ont not in alignments_splice:
            qname_of_map_ont.add(qname_ont)
            continue

        cigar_splice = alignments_splice[qname_ont][5]

        # If preset=splice and inversion is present, `midsv.transform`` will not work, so use preset=map-ont.
        if _has_inversion_in_splice(cigar_splice):
            qname_of_map_ont.add(qname_ont)
            continue

        # only accept single read or inversion reads
        # TODO: 逆位のリードは単純に「３つのリードがある」という条件でいいのか？
        if len(group) == 2 or len(group) >= 4:
            continue
        cigar_ont = group[0][5]

        alignment_length_ont = sam_handler.calculate_alignment_length(cigar_ont)
        alignment_length_splice = sam_handler.calculate_alignment_length(cigar_splice)

        if alignment_length_ont >= alignment_length_splice:
            qname_of_map_ont.add(qname_ont)
    return qname_of_map_ont


def filter_sam_by_preset(sam: Generator[list[str]], qname_of_map_ont: set, preset: str = "map-ont") -> Generator:
    for alignment in sam:
        if alignment[0].startswith("@"):
            yield alignment
        elif preset == "map-ont" and alignment[0] in qname_of_map_ont:
            yield alignment
        elif preset != "map-ont" and alignment[0] not in qname_of_map_ont:
            yield alignment


def transform_to_midsv_format(sam: Generator[list[str]]) -> Generator[list[dict]]:
    for midsv_sample in midsv.transform(sam, midsv=False, cssplit=True, qscore=False, keep=set(["FLAG"])):
        yield midsv_sample


def replace_n_to_d(midsv_sample: Generator[list[dict]], sequence: str) -> Generator[list[dict]]:
    """Replace N with D, but not contiguous from both ends."""
    for samp in midsv_sample:
        cssplits = samp["CSSPLIT"].split(",")

        left_idx_n, right_idx_n = cssplits_handler.find_n_boundaries(cssplits)

        # Replace Ns within the sequence boundaries with the corresponding sequence character
        for j, (cs, seq_char) in enumerate(zip(cssplits, sequence)):
            if left_idx_n < j < right_idx_n and cs == "N":
                cssplits[j] = f"-{seq_char}"

        samp["CSSPLIT"] = ",".join(cssplits)
        yield samp


def convert_flag_to_strand(midsv_sample: Generator[list[str]]) -> Generator[list[dict]]:
    """Convert FLAG to STRAND (+ or -)"""
    REVERSE_STRAND_FLAG = 16 | 2064
    for samp in midsv_sample:
        samp["STRAND"] = "-" if samp["FLAG"] & REVERSE_STRAND_FLAG else "+"
        del samp["FLAG"]
        yield samp


###########################################################
# main
###########################################################


def execute(TEMPDIR: Path | str, FASTA_ALLELES: dict, NAME: str) -> None:
    for allele, sequence in FASTA_ALLELES.items():
        path_output = Path(TEMPDIR, NAME, "midsv", f"{allele}.json")
        if path_output.exists():
            continue
        path_ont = Path(TEMPDIR, NAME, "sam", f"map-ont_{allele}.sam")
        path_splice = Path(TEMPDIR, NAME, "sam", f"splice_{allele}.sam")
        sam_ont = sam_handler.remove_overlapped_reads(list(midsv.read_sam(path_ont)))
        sam_splice = sam_handler.remove_overlapped_reads(list(midsv.read_sam(path_splice)))
        qname_of_map_ont = extract_qname_of_map_ont(sam_ont, sam_splice)
        sam_of_map_ont = filter_sam_by_preset(sam_ont, qname_of_map_ont, preset="map-ont")
        sam_of_splice = filter_sam_by_preset(sam_splice, qname_of_map_ont, preset="splice")
        midsv_chaind = transform_to_midsv_format(chain(sam_of_map_ont, sam_of_splice))
        midsv_sample = replace_n_to_d(midsv_chaind, sequence)
        midsv_sample = convert_flag_to_strand(midsv_sample)
        midsv.write_jsonl(midsv_sample, path_output)



from __future__ import annotations

from pathlib import Path


def create_temporal(TEMPDIR: Path, NAME: str, is_control=False) -> None:
    Path(TEMPDIR, "result").mkdir(parents=True, exist_ok=True)
    if is_control:
        SUBDIRS = ["fasta", "sam", "midsv", "mutation_loci", "clustering"]
    else:
        SUBDIRS = [
            "fasta",
            "sam",
            "midsv",
            "mutation_loci",
            "knockin_loci",
            "classification",
            "clustering",
            "consensus",
        ]
    for subdir in SUBDIRS:
        Path(TEMPDIR, NAME, subdir).mkdir(parents=True, exist_ok=True)


def create_report(TEMPDIR: Path, NAME: str, is_control=False) -> None:
    if is_control:
        Path(TEMPDIR, "report", "BAM", NAME).mkdir(parents=True, exist_ok=True)
        return
    SUBDIRS_REPORT = ["HTML", "FASTA", "BAM", "MUTATION_INFO", ".igvjs"]
    for reportdir in SUBDIRS_REPORT:
        if reportdir == "MUTATION_INFO":
            Path(TEMPDIR, "report", reportdir).mkdir(parents=True, exist_ok=True)
        else:
            Path(TEMPDIR, "report", reportdir, NAME).mkdir(parents=True, exist_ok=True)



from __future__ import annotations

import shutil
from pathlib import Path

import midsv
import pandas as pd
import plotly.express as px

from DAJIN2.utils.config import DAJIN_RESULTS_DIR, TEMP_ROOT_DIR


def extract_all_info(batch_directory: Path) -> pd.DataFrame:
    df_results = pd.DataFrame()
    colum = ["SAMPLE", "QNAME", "NAME", "READNUM", "PERCENT", "LABEL", "ALLELE"]
    for path_result in sorted(batch_directory.iterdir()):
        sample_name = path_result.stem
        result_jsonl = midsv.read_jsonl(path_result)
        df_clust_sample = pd.DataFrame(result_jsonl)
        df_clust_sample["SAMPLE"] = sample_name
        df_clust_sample = df_clust_sample[colum]
        df_results = pd.concat([df_results, df_clust_sample])
    return df_results


def summarize_info(df_all: pd.DataFrame) -> pd.DataFrame:
    df_summary = df_all.drop(columns=["QNAME", "LABEL", "ALLELE"]).drop_duplicates()
    return df_summary


def output_plot(df_summary: pd.DataFrame, report_directory: Path):
    df_plot = df_summary.copy()
    name = df_plot.NAME.tolist()
    names = [n.split("_") for n in name]
    alleletype = [n[1] + " " + n[2] for n in names]
    df_plot["ALLELETYPE"] = alleletype
    fig = px.bar(
        df_plot,
        x="SAMPLE",
        y="PERCENT",
        color="ALLELETYPE",
        text="PERCENT",
        labels={"SAMPLE": "Samples", "PERCENT": "% of reads", "ALLELETYPE": "Alelle type"},
    )
    fig.update_traces(textposition="inside", cliponaxis=False)
    fig.update_xaxes(categoryorder="category ascending")
    output_filename = str(Path(report_directory, "read_plot"))
    fig.write_html(f"{output_filename}.html")
    # if kaleido is installed, output a pdf
    try:
        fig.write_image(f"{output_filename}.pdf")
    except ValueError:
        pass


def report(name: str):
    report_directory = Path(DAJIN_RESULTS_DIR, name)
    report_directory.mkdir(exist_ok=True, parents=True)
    shutil.copytree(Path(TEMP_ROOT_DIR, name, "report"), report_directory, dirs_exist_ok=True)
    df_all = extract_all_info(Path(TEMP_ROOT_DIR, name, "result"))
    df_all.to_csv(Path(report_directory, "read_all.csv"), index=False)
    df_summary = summarize_info(df_all)
    df_summary.to_csv(Path(report_directory, "read_summary.csv"), index=False)
    output_plot(df_summary, report_directory)
